import logging
import urllib.parse
from typing import Any, List, Optional

import requests
from retry.api import retry_call

from .__version__ import __version__
from .errors import SnykHTTPError, SnykNotImplementedError
from .managers import Manager
from .models import Organization, Project
from .utils import cleanup_path

logger = logging.getLogger(__name__)


class SnykClient(object):
    API_URL = "https://api.snyk.io/v1"
    REST_API_URL = "https://api.snyk.io/rest"
    USER_AGENT = "pysnyk/%s" % __version__

    def __init__(
        self,
        token: str,
        url: Optional[str] = None,
        user_agent: Optional[str] = USER_AGENT,
        debug: bool = False,
        tries: int = 1,
        delay: int = 1,
        backoff: int = 2,
        verify: bool = True,
        version: Optional[str] = None,
    ):
        self.api_token = token
        self.api_url = url or self.API_URL
        self.api_headers = {
            "Authorization": "token %s" % self.api_token,
            "User-Agent": user_agent,
        }
        self.api_post_headers = self.api_headers
        self.api_post_headers["Content-Type"] = "application/json"
        self.tries = tries
        self.backoff = backoff
        self.delay = delay
        self.verify = verify
        self.version = version

        # Ensure we don't have a trailing /
        if self.api_url[-1] == "/":
            self.api_url = self.api_url.rstrip("/")

        if debug:
            logging.basicConfig(level=logging.DEBUG)

    def request(
        self,
        method,
        url: str,
        headers: object,
        params: object = None,
        json: object = None,
    ) -> requests.Response:

        if params and json:
            resp = method(
                url, headers=headers, params=params, json=json, verify=self.verify
            )
        elif params and not json:
            resp = method(url, headers=headers, params=params, verify=self.verify)
        elif json and not params:
            resp = method(url, headers=headers, json=json, verify=self.verify)
        else:
            resp = method(url, headers=headers, verify=self.verify)

        if not resp or resp.status_code >= requests.codes.server_error:
            logger.warning(f"Retrying: {resp.text}")
            raise SnykHTTPError(resp)
        return resp

    def post(self, path: str, body: Any, headers: dict = {}) -> requests.Response:
        url = f"{self.api_url}/{path}"
        logger.debug(f"POST: {url}")

        resp = retry_call(
            self.request,
            fargs=[requests.post, url],
            fkwargs={"json": body, "headers": {**self.api_post_headers, **headers}},
            tries=self.tries,
            delay=self.delay,
            backoff=self.backoff,
            exceptions=SnykHTTPError,
            logger=logger,
        )

        if not resp.ok:
            logger.error(resp.text)
            raise SnykHTTPError(resp)

        return resp

    def put(self, path: str, body: Any, headers: dict = {}) -> requests.Response:
        url = "%s/%s" % (self.api_url, path)
        logger.debug("PUT: %s" % url)

        resp = retry_call(
            self.request,
            fargs=[requests.put, url],
            fkwargs={"json": body, "headers": {**self.api_post_headers, **headers}},
            tries=self.tries,
            delay=self.delay,
            backoff=self.backoff,
            logger=logger,
        )
        if not resp.ok:
            logger.error(resp.text)
            raise SnykHTTPError(resp)

        return resp

    def get(
        self,
        path: str,
        params: dict = None,
        version: str = None,
        exclude_version: bool = False,
    ) -> requests.Response:
        """
        Rest (formerly v3) Compatible Snyk Client, assumes the presence of Version, either set in the client
        or called in this method means that we're talking to a rest API endpoint and will ensure the
        params are encoded properly with the version.

        Since certain endpoints can exist only in certain versions, being able to override the
        client version with each GET is necessary

        Returns a standard requests Response object
        """

        path = cleanup_path(path)
        if version:
            # When calling a "next page" link, it fails if a version parameter is appended on to the URL - this is a
            # workaround to prevent that from happening...
            if exclude_version:
                url = f"{self.REST_API_URL}/{path}"
            else:
                url = f"{self.REST_API_URL}/{path}?version={version}"
        else:
            url = f"{self.api_url}/{path}"

        if params or self.version:

            if not params:
                params = {}

            # we use the presence of version to determine if we are REST or not
            if "version" not in params.keys() and self.version:
                params["version"] = version or self.version

            # Python Bools are True/False, JS Bools are true/false
            # Snyk REST API is strictly case sensitive at the moment

            for k, v in params.items():
                if isinstance(v, bool):
                    params[k] = str(v).lower()

            debug_url = f"{url}&{urllib.parse.urlencode(params)}"
            fkwargs = {"headers": self.api_headers, "params": params}
        else:
            debug_url = url
            fkwargs = {"headers": self.api_headers}

        logger.debug(f"GET: {debug_url}")

        resp = retry_call(
            self.request,
            fargs=[requests.get, url],
            fkwargs=fkwargs,
            tries=self.tries,
            delay=self.delay,
            backoff=self.backoff,
            logger=logger,
        )
        if not resp.ok:
            logger.error(resp.text)
            raise SnykHTTPError(resp)

        return resp

    def delete(self, path: str) -> requests.Response:
        url = f"{self.api_url}/{path}"
        logger.debug(f"DELETE: {url}")

        resp = retry_call(
            self.request,
            fargs=[requests.delete, url],
            fkwargs={"headers": self.api_headers},
            tries=self.tries,
            delay=self.delay,
            backoff=self.backoff,
            logger=logger,
        )
        if not resp.ok:
            logger.error(resp.text)
            raise SnykHTTPError(resp)

        return resp

    def get_rest_pages(self, path: str, params: dict = {}) -> List:
        """
        Helper function to collect paginated responses from the rest API into a single
        list.

        This collects the "data" list from the first reponse and then appends the
        any further "data" lists if a next link is found in the links field.
        """

        # this is a raw primative but a higher level module might want something that does an
        # arbitrary path + origin=foo + limit=100 url construction instead before being sent here

        limit = params["limit"]

        data = list()

        page = self.get(path, params).json()

        data.extend(page["data"])

        while "next" in page["links"].keys():
            logger.debug(
                f"GET_REST_PAGES: Another link exists: {page['links']['next']}"
            )

            next_url = urllib.parse.urlsplit(page["links"]["next"])
            query = urllib.parse.parse_qs(next_url.query)

            for k, v in query.items():
                params[k] = v

            params["limit"] = limit

            page = self.get(next_url.path, params).json()

            data.extend(page["data"])

            logger.debug(
                f"GET_REST_PAGES: Added another {len(page['data'])} items to the response"
            )

        return data

    # alias for backwards compatibility where V3 was the old name
    get_v3_pages = get_rest_pages

    @property
    def organizations(self) -> Manager:
        return Manager.factory(Organization, self)

    @property
    def projects(self) -> Manager:
        return Manager.factory(Project, self)

    # https://snyk.docs.apiary.io/#reference/general/the-api-details/get-notification-settings
    # https://snyk.docs.apiary.io/#reference/users/user-notification-settings/modify-notification-settings
    def notification_settings(self):
        raise SnykNotImplementedError  # pragma: no cover

    # https://snyk.docs.apiary.io/#reference/groups/organisations-in-groups/create-a-new-organisation-in-the-group
    # https://snyk.docs.apiary.io/#reference/0/list-members-in-a-group/list-all-members-in-a-group
    # https://snyk.docs.apiary.io/#reference/0/members-in-an-organisation-of-a-group/add-a-member-to-an-organisation-from-another-organisation-in-the-group
    def groups(self):
        raise SnykNotImplementedError  # pragma: no cover

    # https://snyk.docs.apiary.io/#reference/reporting-api/issues/get-list-of-issues
    def issues(self):
        raise SnykNotImplementedError  # pragma: no cover



import os
import re

import pytest  # type: ignore

from snyk import SnykClient
from snyk.__version__ import __version__
from snyk.errors import SnykError, SnykNotFoundError
from snyk.models import Organization, Project
from snyk.utils import load_test_data

TEST_DATA = os.path.join(os.path.dirname(__file__), "test_data")

REST_ORG = "39ddc762-b1b9-41ce-ab42-defbe4575bd6"
REST_URL = "https://api.snyk.io/rest"
REST_VERSION = "2022-02-16~experimental"

V3_ORG = "39ddc762-b1b9-41ce-ab42-defbe4575bd6"
V3_URL = "https://api.snyk.io/v3"
V3_VERSION = "2022-02-16~experimental"


class TestSnykClient(object):
    @pytest.fixture
    def client(self):
        return SnykClient("token")

    def test_default_api_url(self, client):
        assert client.api_url == "https://api.snyk.io/v1"

    def test_overriding_api_url(self):
        url = "https://api.notsnyk.io/v1"
        client = SnykClient("token", url)
        assert client.api_url == url

    def test_token_added_to_headers(self, client):
        assert client.api_headers["Authorization"] == "token token"

    def test_user_agent_added_to_headers(self, client):
        assert client.api_headers["User-Agent"] == "pysnyk/%s" % __version__

    def test_overriding_user_agent(self):
        ua = "test"
        client = SnykClient("token", user_agent=ua)
        assert client.api_headers["User-Agent"] == ua

    def test_token_added_to_post_headers(self, client):
        assert client.api_post_headers["Authorization"] == "token token"

    def test_post_headers_use_correct_mimetype(self, client):
        assert client.api_post_headers["Content-Type"] == "application/json"

    def test_get_sends_request_to_snyk(self, requests_mock, client):
        requests_mock.get("https://api.snyk.io/v1/sample", text="pong")
        assert client.get("sample")

    def test_put_sends_request_to_snyk(self, requests_mock, client):
        requests_mock.put("https://api.snyk.io/v1/sample", text="pong")
        assert client.put("sample", {})

    def test_delete_sends_request_to_snyk(self, requests_mock, client):
        requests_mock.delete("https://api.snyk.io/v1/sample")
        assert client.delete("sample")

    def test_post_sends_request_to_snyk(self, requests_mock, client):
        requests_mock.post("https://api.snyk.io/v1/sample")
        assert client.post("sample", {})
        assert requests_mock.call_count == 1

    def test_post_raises_error(self, requests_mock, client):
        requests_mock.post("https://api.snyk.io/v1/sample", status_code=500, json={})
        with pytest.raises(SnykError):
            client.post("sample", {})
        assert requests_mock.call_count == 1

    def test_put_retries_and_raises_error(self, requests_mock, client):
        requests_mock.put("https://api.snyk.io/v1/sample", status_code=500, json={})
        client = SnykClient("token", tries=4, delay=0, backoff=2)
        with pytest.raises(SnykError):
            client.put("sample", {})
        assert requests_mock.call_count == 4

    def test_delete_retries_and_raises_error(self, requests_mock, client):
        requests_mock.delete("https://api.snyk.io/v1/sample", status_code=500, json={})
        client = SnykClient("token", tries=4, delay=0, backoff=2)
        with pytest.raises(SnykError):
            client.delete("sample")
        assert requests_mock.call_count == 4

    def test_get_retries_and_raises_error(self, requests_mock, client):
        requests_mock.get("https://api.snyk.io/v1/sample", status_code=500, json={})
        client = SnykClient("token", tries=4, delay=0, backoff=2)
        with pytest.raises(SnykError):
            client.get("sample")
        assert requests_mock.call_count == 4

    def test_post_retries_and_raises_error(self, requests_mock, client):
        requests_mock.post("https://api.snyk.io/v1/sample", status_code=500, json={})
        client = SnykClient("token", tries=4, delay=0, backoff=2)
        with pytest.raises(SnykError):
            client.post("sample", {})
        assert requests_mock.call_count == 4

    def test_put_raises_error(self, requests_mock, client):
        requests_mock.put("https://api.snyk.io/v1/sample", status_code=500, json={})
        with pytest.raises(SnykError):
            client.put("sample", {})
        assert requests_mock.call_count == 1

    def test_delete_raises_error(self, requests_mock, client):
        requests_mock.delete("https://api.snyk.io/v1/sample", status_code=500, json={})
        with pytest.raises(SnykError):
            client.delete("sample")
        assert requests_mock.call_count == 1

    def test_get_raises_error(self, requests_mock, client):
        requests_mock.get("https://api.snyk.io/v1/sample", status_code=500, json={})
        with pytest.raises(SnykError):
            client.get("sample")
        assert requests_mock.call_count == 1

    def test_empty_organizations(self, requests_mock, client):
        requests_mock.get("https://api.snyk.io/v1/orgs", json={})
        assert [] == client.organizations.all()

    @pytest.fixture
    def organizations(self):
        return load_test_data(TEST_DATA, "organizations")

    @pytest.fixture
    def projects(self):
        return load_test_data(TEST_DATA, "projects")

    def test_loads_organizations(self, requests_mock, client, organizations):
        requests_mock.get("https://api.snyk.io/v1/orgs", json=organizations)
        assert len(client.organizations.all()) == 2

    def test_first_organizations(self, requests_mock, client, organizations):
        requests_mock.get("https://api.snyk.io/v1/orgs", json=organizations)
        org = client.organizations.first()
        assert "defaultOrg" == org.name

    def test_first_organizations_on_empty(self, requests_mock, client):
        requests_mock.get("https://api.snyk.io/v1/orgs", json={})
        with pytest.raises(SnykNotFoundError):
            client.organizations.first()

    def test_filter_organizations(self, requests_mock, client, organizations):
        requests_mock.get("https://api.snyk.io/v1/orgs", json=organizations)
        assert 1 == len(client.organizations.filter(name="defaultOrg"))

    def test_filter_organizations_empty(self, requests_mock, client, organizations):
        requests_mock.get("https://api.snyk.io/v1/orgs", json=organizations)
        assert [] == client.organizations.filter(name="not present")

    def test_loads_organization(self, requests_mock, client, organizations):
        key = organizations["orgs"][0]["id"]
        requests_mock.get("https://api.snyk.io/v1/orgs", json=organizations)
        org = client.organizations.get(key)
        assert "defaultOrg" == org.name

    def test_non_existent_organization(self, requests_mock, client, organizations):
        requests_mock.get("https://api.snyk.io/v1/orgs", json=organizations)
        with pytest.raises(SnykNotFoundError):
            client.organizations.get("not-present")

    def test_organization_type(self, requests_mock, client, organizations):
        requests_mock.get("https://api.snyk.io/v1/orgs", json=organizations)
        assert all(type(x) is Organization for x in client.organizations.all())

    def test_organization_attributes(self, requests_mock, client, organizations):
        requests_mock.get("https://api.snyk.io/v1/orgs", json=organizations)
        assert client.organizations.first().name == "defaultOrg"

    def test_organization_load_group(self, requests_mock, client, organizations):
        requests_mock.get("https://api.snyk.io/v1/orgs", json=organizations)
        assert client.organizations.all()[1].group.name == "ACME Inc."

    def test_empty_projects(self, requests_mock, client, organizations):
        requests_mock.get("https://api.snyk.io/v1/orgs", json=organizations)
        matcher = re.compile("projects.*$")
        requests_mock.get(matcher, json={})
        assert [] == client.projects.all()

    def test_projects(self, requests_mock, client, organizations, projects):
        requests_mock.get("https://api.snyk.io/v1/orgs", json=organizations)
        matcher = re.compile("projects.*$")
        requests_mock.get(matcher, json=projects)
        assert len(client.projects.all()) == 2
        assert all(type(x) is Project for x in client.projects.all())

    def test_project(self, requests_mock, client, organizations, projects):
        requests_mock.get("https://api.snyk.io/v1/orgs", json=organizations)
        matcher = re.compile("projects.*$")
        requests_mock.get(matcher, json=projects)
        assert (
            "testing-new-name"
            == client.projects.get("f9fec29a-d288-40d9-a019-cedf825e6efb").name
        )

    def test_non_existent_project(self, requests_mock, client, organizations, projects):
        requests_mock.get("https://api.snyk.io/v1/orgs", json=organizations)
        matcher = re.compile("projects.*$")
        requests_mock.get(matcher, json=projects)
        with pytest.raises(SnykNotFoundError):
            client.projects.get("not-present")

    @pytest.fixture
    def rest_client(self):
        return SnykClient(
            "token", version="2022-02-16~experimental", url="https://api.snyk.io/rest"
        )

    @pytest.fixture
    def v3_client(self):
        return SnykClient(
            "token", version="2022-02-16~experimental", url="https://api.snyk.io/v3"
        )

    @pytest.fixture
    def v3_groups(self):
        return load_test_data(TEST_DATA, "v3_groups")

    @pytest.fixture
    def v3_targets_page1(self):
        return load_test_data(TEST_DATA, "v3_targets_page1")

    @pytest.fixture
    def v3_targets_page2(self):
        return load_test_data(TEST_DATA, "v3_targets_page2")

    @pytest.fixture
    def v3_targets_page3(self):
        return load_test_data(TEST_DATA, "v3_targets_page3")

    @pytest.fixture
    def rest_groups(self):
        return load_test_data(TEST_DATA, "rest_groups")

    @pytest.fixture
    def rest_targets_page1(self):
        return load_test_data(TEST_DATA, "rest_targets_page1")

    @pytest.fixture
    def rest_targets_page2(self):
        return load_test_data(TEST_DATA, "rest_targets_page2")

    @pytest.fixture
    def rest_targets_page3(self):
        return load_test_data(TEST_DATA, "rest_targets_page3")

    def test_v3get(self, requests_mock, v3_client, v3_targets_page1):
        requests_mock.get(
            f"{V3_URL}/orgs/{V3_ORG}/targets?limit=10&version={V3_VERSION}",
            json=v3_targets_page1,
        )
        t_params = {"limit": 10}

        targets = v3_client.get(f"orgs/{V3_ORG}/targets", t_params).json()

        assert len(targets["data"]) == 10

    def test_get_v3_pages(
        self,
        requests_mock,
        v3_client,
        v3_targets_page1,
        v3_targets_page2,
        v3_targets_page3,
    ):
        requests_mock.get(
            f"{V3_URL}/orgs/{V3_ORG}/targets?limit=10&version={V3_VERSION}",
            json=v3_targets_page1,
        )
        requests_mock.get(
            f"{V3_URL}/orgs/{V3_ORG}/targets?limit=10&version={V3_VERSION}&excludeEmpty=true&starting_after=v1.eyJpZCI6IjMyODE4ODAifQ%3D%3D",
            json=v3_targets_page2,
        )
        requests_mock.get(
            f"{V3_URL}/orgs/{V3_ORG}/targets?limit=10&version={V3_VERSION}&excludeEmpty=true&starting_after=v1.eyJpZCI6IjI5MTk1NjgifQ%3D%3D",
            json=v3_targets_page3,
        )
        t_params = {"limit": 10}

        data = v3_client.get_v3_pages(f"orgs/{V3_ORG}/targets", t_params)

        assert len(data) == 30

    def test_rest_get(self, requests_mock, rest_client, rest_targets_page1):
        requests_mock.get(
            f"{REST_URL}/orgs/{REST_ORG}/targets?limit=10&version={REST_VERSION}",
            json=rest_targets_page1,
        )
        t_params = {"limit": 10}

        targets = rest_client.get(f"orgs/{REST_ORG}/targets", t_params).json()

        assert len(targets["data"]) == 10

    def test_get_rest_pages(
        self,
        requests_mock,
        rest_client,
        rest_targets_page1,
        rest_targets_page2,
        rest_targets_page3,
    ):
        requests_mock.get(
            f"{REST_URL}/orgs/{REST_ORG}/targets?limit=10&version={REST_VERSION}",
            json=rest_targets_page1,
        )
        requests_mock.get(
            f"{REST_URL}/orgs/{REST_ORG}/targets?limit=10&version={REST_VERSION}&excludeEmpty=true&starting_after=v1.eyJpZCI6IjMyODE4ODAifQ%3D%3D",
            json=rest_targets_page2,
        )
        requests_mock.get(
            f"{REST_URL}/orgs/{REST_ORG}/targets?limit=10&version={REST_VERSION}&excludeEmpty=true&starting_after=v1.eyJpZCI6IjI5MTk1NjgifQ%3D%3D",
            json=rest_targets_page3,
        )
        t_params = {"limit": 10}

        data = rest_client.get_rest_pages(f"orgs/{V3_ORG}/targets", t_params)

        assert len(data) == 30



from importlib_metadata import version  # type: ignore

__title__ = "snyk"
__description__ = "An API client for the Snyk API."
__url__ = "https://snyk.docs.apiary.io"
__license__ = "MIT"
__version__ = version("pysnyk")



import json
import logging
from itertools import chain


def snake_to_camel(word):
    return lower_case_first_letter(
        "".join(x.capitalize() or "_" for x in word.split("_"))
    )


def lower_case_first_letter(word):
    return word[:1].lower() + word[1:] if word else ""


def flat_map(fn, *args):
    mapped = map(fn, *args)
    return list(chain(*mapped))


def format_package(pkg):
    return "{name}@{version}".format(name=pkg.name, version=pkg.version or "*")


def cleanup_path(path: str) -> str:
    """
    Strings '/' from the start and end of strings if present to ensure that a '//' doesn't
    occur in an API request due to copy/paste error
    """

    if path[0] == "/":
        path = path[1:]
    if path[-1] == "/":
        path = path.rstrip("/")

    return path


def load_test_data(test_dir: str, test_name: str) -> dict:
    """
    Returns the contents of a json file at location of:
    test_dir/test_name.json

    This is meant to keep large amounts of json needed for testing outside of
    the tests themselves and as the actual json responses from the API
    """
    test_file = f"{test_dir}/{test_name}.json"
    with open(test_file, "r") as the_file:
        data = the_file.read()
    return json.loads(data)
